/* cknife, a cryptographic "swiss army knife" utility
 * version 2.2 by Pawel Krawczyk <kravietz at echelon dot pl>
 *
 * Features:
 * 	(e) encrypts arbitrary data using 3DES-CBC with DESX style
 * 	    whitening and HMAC-SHA384-160 integrity protection
 * 	(d) decrypts resulting files with integrity verification
 * 	(k) calculates HMAC-SHA384-160 of given file and key
 * 	(r) calculates SHA384-160 for given file
 * The (x) characters denote command line options.
 *
 * Output file format:
 * offset length  description
 * 0      20      HMAC-SHA384-160
 * 20      4      length in bytes of "data" field
 * 24     16      initialisation vector for 3DES-CBC
 * 39   n*16       data in 8 byte, padded blocks
 *
 * Encryption facts in cknife:
 * 	(1) 3DES-CBC with random IV is used for encryption
 * 	(2) IV is stored in the output file
 * 	(3) for every block of input data a HMAC is calculated
 * 	    using HMAC-SHA384 method
 * 	(4) every resulting HMAC is hashed with global MAC
 * 	    using SHA384
 * 	(5) first 160 bits of resulting MAC are stored in the output file
 * 	(6) plaintext length is stored in the output file in network format
 * 	(7) password for encryption and HMAC is passed through SHA384,
 * 	    first 192 bits are taken for HMAC and 3DES, next 128 bits
 * 	    are taken for inner and outer 3DES whitening
 * 	(8) block length for CBC is 16 bytes
 * 	(9) plaintext blocks shorter than input blocks are padded
 * 	    with random data
 *     (10) whitening is done around 3DES-EDE, not inside, to be
 *          precise: C = Kc XOR 3DES(Kp XOR P)
 *
 * Redistribution of this file is permitted under the GNU Public License.
 */ 

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "sha2.h"
#include "des.h"

#ifndef DEBUG
# define DEBUG	0
#endif


/* Modes of operation: */
#define CKNIFE_ENCRYPT		0x01	/* 3DES encryption */
#define	CKNIFE_DECRYPT		0x02	/* 3DES decryption */
#define	CKNIFE_SHA384		0x03	/* SHA384-160 hash */
#define	CKNIFE_HMAC_SHA384	0x04	/* HMAC-SHA384-160 hash */

int main(int argc, char **argv) {
	int fdin=-1, fdout=-1;
	unsigned char *ifile=NULL, *ofile=NULL;
	unsigned long length = 0L;
	int left = 0;
#define	BLOCK_SIZE	16	/* Work block size in bytes for CBC */
#define	MAC_SIZE	20	/* Effective MAC length in bytes */
#define SCH_SIZE	8*16	/* DES key schedule size */
	unsigned char in[BLOCK_SIZE],
		out[BLOCK_SIZE]; /* in/out buffers */
	unsigned char ipad[BLOCK_SIZE],
		opad[BLOCK_SIZE]; /* inner/outer HMAC pads */
	unsigned char ks1[SCH_SIZE], ks2[SCH_SIZE],
		ks3[SCH_SIZE]; /* three DES key schedules for 3DES */
	unsigned char k[8]; /* temporary buffer for DES key */
	unsigned char ivec[BLOCK_SIZE];	/* CBC vector */
	unsigned char iwhite[BLOCK_SIZE],
		owhite[BLOCK_SIZE]; /* whitening keys */
	SHA384_CTX c, cglobal;	/* SHA384 context for local and global MAC */
	unsigned char md[SHA384_DIGEST_LENGTH], resmd[MAC_SIZE];
	unsigned char *pass=NULL;
	int r=-1, w=-1, mode=-1, i=-1;

	if(argc < 2) {
		printf("cknife 2.2 by Pawel Krawczyk <kravietz@echelon.pl>\n");
		printf("Usage: cknife [e|d] secret [in [out]] (3DES-CBC w/HMAC-SHA384)\n");
		printf("       cknife k secret [in]           (HMAC-SHA384)\n");
		printf("       cknife r [in]                  (SHA384)\n");
		return 1;
	}
	switch(*argv[1]) {
		case 'k':
			mode = CKNIFE_HMAC_SHA384;
			if(argc < 3) {
				printf("usage: cknife k secret [in]\n");
				return 1;
			}
			break;
		case 'r':
			mode =  CKNIFE_SHA384;
			if(argc < 2) {
				printf("usage: cknife r [in]\n");
				return 1;
			}
			break;
		case 'e':
		case 'd':
			if(argc < 3) {
				printf("usage: cknife e/d secret [in [out]]\n");
				return 1;
			}
			mode = (*argv[1] == 'e') ? CKNIFE_ENCRYPT : CKNIFE_DECRYPT;
			break;
		default:
			printf("invalid command `%c'\n", *argv[1]);
			return 2;

	};

	if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT 
	                       || mode == CKNIFE_HMAC_SHA384) {
		pass = strdup(argv[2]);
		memset(argv[2], 0, strlen(argv[2]));
		ifile = argv[3];
		ofile = argv[4];
		/* HMAC pad tables initialisation (RFC 2104)
		 *
		 *  ipad = the byte 0x36 repeated B times
		 *  opad = the byte 0x5C repeated B times
		 *  B=20, L=20, K=SHA384(password)
		 */
		memset(ipad, 0, sizeof(ipad));
		memset(opad, 0, sizeof(opad));

#if DEBUG < 2
		/* password gets converted into 384 bit hash */
		SHA384_Init(&c);
		SHA384_Update(&c, pass, strlen(pass));
		SHA384_Final(md, &c);
		memset(pass, '[', strlen(pass));
		free(pass);

		/* first 16 bytes of hashed password go to HMAC */
		memcpy(ipad, md, sizeof(ipad));
		memcpy(opad, md, sizeof(ipad));
		for(i=0; i<sizeof(ipad); i++) ipad[i] ^= 0x36;
		for(i=0; i<sizeof(opad); i++) opad[i] ^= 0x5c;
#if DEBUG > 0
		printf("\nHMAC ipad: ");
		for(i=0; i<sizeof(ipad); i++) printf("%02x", ipad[i]);
		printf("\nHMAC opad: ");
		for(i=0; i<sizeof(opad); i++) printf("%02x", opad[i]);
#endif
		/* Teraz:
		 *  ipad = K XOR ipad
		 *  opad = K XOR opad
		 */

#else
		memset(&md, 0, sizeof(md));
#endif
	}
	if(mode == CKNIFE_SHA384) {
		ifile = argv[2];
		ofile = argv[3];
	}

	/* file manipulation */
	if(ifile == NULL || *ifile == '-') {
		fdin = 0;
	} else {
		fdin = open(ifile, O_RDONLY);
		if( fdin < 0 ) {
			perror("input file");
			return 3;
		}
	}
	if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT) {
		if(ofile == NULL || *ofile == '-') {
			fdout = 1;
		} else {
			fdout = open(ofile, O_WRONLY|O_CREAT|O_TRUNC,
					S_IRUSR|S_IWUSR);
			if( fdout < 0 ) {
				perror("output file");
				return 4;
			}
		}
	}

	
	if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT) {
		/* Keys assigned from hashed password
		 * DES keys:
		 *  k1: bytes 0-7
		 *  k2: bytes 8-15
		 *  k3: bytes 16-23
		 * whitening keys:
		 *  inner white: bytes 24-31
		 *  outer white: bytes 32-40
		 */
		desinit(0);

		memcpy(&k, md, 8);
		setkey(ks1, k);
#if DEBUG > 0
		printf("\nDES k1: ");
		for(i=0; i<8; i++) printf("%02x ", k[i]);
#endif
		memcpy(&k, (md+8), 8); 
		setkey(ks2, k);
#if DEBUG > 0
		printf("\nDES k2: ");
		for(i=0; i<8; i++) printf("%02x ", k[i]);
#endif
		memcpy(&k, (md+16), 8);
		setkey(ks3, k);
#if DEBUG > 0
		printf("\nDES k3: ");
		for(i=0; i<8; i++) printf("%02x ", k[i]);
#endif
		memset(&k, 0, sizeof(k));

		/* whitening */
		memcpy(iwhite, (md+24), 8);
		memcpy(owhite, (md+32), 8);

		memset(&md, 0, sizeof(md));
	}

	if(mode == CKNIFE_ENCRYPT) {
		int fdrand;
		/* we create placeholder for MAC in the output file */
		memset(md, 0, sizeof(md));
		w = write(fdout, md, 20);
		if( w < 20) {
			perror("write md");
			return 5;
		}
		/* placeholder for length */
		length = 0L;
		w = write(fdout, &length, sizeof(length));
		if( w < sizeof(length)) {
			perror("write length");
			return 12;
		}
		/* we create random IV */
#if DEBUG < 2
		fdrand = open("/dev/urandom", O_RDONLY);
		if( fdrand < 0) {
			perror("open urandom");
			return 7;
		}
		r = read(fdrand, &ivec, sizeof(ivec));
		if( r < sizeof(ivec) ) {
			perror("read urandom");
			return 8;
		}
		close(fdrand);
#else
		memset(&ivec, 0, sizeof(ivec));
#endif
		/* we store random IV */
		w = write(fdout, &ivec, sizeof(ivec));
		if( w < sizeof(ivec)) {
			perror("write ivec");
			return 11;
		}
	}
	if(mode == CKNIFE_DECRYPT) {
		/* we read MAC for the encrypted plaintext */
		r = read(fdin, resmd, MAC_SIZE);
		if( r < MAC_SIZE) {
			perror("read resmd");
			return 9;
		}
		/* we read length */
		r = read(fdin, &length, sizeof(length));
		if(r  < sizeof(length)) {
			perror("read length");
			return 13;
		}
		length = ntohl(length);
#if DEBUG > 0
		printf("\nOriginal MAC: ");
		for(i=0; i<20; i++) printf("%02x ", resmd[i]);
#endif
		/* we read random IV */
		r = read(fdin, &ivec, sizeof(ivec));
		if( r < sizeof(ivec) ) {
			perror("read ivec");
			return 10;
		}

	}

	/* initialise the global MAC */
	SHA384_Init(&cglobal);

	/* actual encryption/decryption with MAC calculations */
	if(mode == CKNIFE_ENCRYPT)
		left = 0;
	while(1) {
		unsigned char *data;

		/* read plain- or ciphertext */
		memset(&in, 0, sizeof(in));
		r = read(fdin, in, sizeof(in));
		if(r < 0 ) {
			perror("read data");
			return 11;
		}
		if(r == 0) break;

		if(mode == CKNIFE_DECRYPT && r < sizeof(in)) {
			printf("error: truncated ciphertext!\n");
			printf("error: read %u bytes, should be %u\n",
					r, sizeof(in));
			return 26;
		}

		if(mode == CKNIFE_ENCRYPT) {
			length += r;
		}
		if(mode == CKNIFE_DECRYPT) {
			if(length < sizeof(in)) {
				left = length;
			} else {
				length -= r;
			}
		}

#if DEBUG > 0
		if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT) {
			printf("\nIV1: ");
			for(i=0; i<8; i++) printf("%02x ", ivec[i]);
		}
#endif

		if(mode == CKNIFE_ENCRYPT) {
			/* Random padding, so that we don't have known
			 * plaintext of 0s in the last block,
			 * as the original plaintext length is known
			 */
			left = r % BLOCK_SIZE;
			if(left != 0) {
				int rfd = -1, rr = -1;

				rfd = open("/dev/urandom", O_RDONLY);
				if(rfd < 0) {
					perror("open random pad");
					return 23;
				}
#if DEBUG > 0
				printf("\npadding %d", (sizeof(out)-left));
#endif
				rr = read(rfd, in+left, (sizeof(in)-left) );
				if(rr < (sizeof(in)-left) ) {
					perror("read random pad");
					return 24;
				}
				close(rfd);
			}
		}

		/* encrypt/decrypt */
		memset(&out, 0, sizeof(out));
		if(mode == CKNIFE_ENCRYPT) {
			for(i=0; i<sizeof(in); i++) out[i] = in[i] ^ ivec[i];
			for(i=0; i<sizeof(in); i++) out[i] ^= iwhite[i];
			endes(ks1, out);
			dedes(ks2, out);
			endes(ks3, out);
			endes(ks1, out+8);
			dedes(ks2, out+8);
			endes(ks3, out+8);
			for(i=0; i<sizeof(in); i++) out[i] ^= owhite[i];
			memcpy(ivec, out, sizeof(ivec));
		}
		if(mode == CKNIFE_DECRYPT) {
			for(i=0; i<sizeof(in); i++) out[i] = in[i] ^ owhite[i];
			dedes(ks3, out);
			endes(ks2, out);
			dedes(ks1, out);
			dedes(ks3, out+8);
			endes(ks2, out+8);
			dedes(ks1, out+8);
			for(i=0; i<sizeof(in); i++) out[i] ^= iwhite[i];
			for(i=0; i<sizeof(in); i++) out[i] ^= ivec[i];
			memcpy(ivec, in, sizeof(ivec));
		}


#if DEBUG > 0
		if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT) {
			printf("\nInput: ");
			for(i=0; i<sizeof(in); i++) printf("%02x ", in[i]);
			printf("\nOutput: ");
			for(i=0; i<sizeof(out); i++) printf("%02x ", out[i]);
			printf("\nIV2: ");
			for(i=0; i<8; i++) printf("%02x ", ivec[i]);
		}
#endif

		if(mode == CKNIFE_ENCRYPT) {
			w = write(fdout, out, sizeof(out));
			if( w < sizeof(out)) {
				perror("writing ciphertext");
				return(24);
			}
		}
		if(mode == CKNIFE_DECRYPT) {
			w = write(fdout, out,
					(left == 0 ? sizeof(out) : left) );
			if( w < (left == 0 ? sizeof(out) : left) ) {
				perror("writing plaintext");
				return(15);
			}
		}

		/* HMAC-SHA384, local (per-block) MAC calculation
		 * it consists of inner and outer mac, the inner
		 * mac is calculated for the whole 20 bytes buffer;
		 * when data is shorter, the rest is already padded
		 * with zeros
		 */
		data = in;
		if(mode == CKNIFE_DECRYPT) data = out;
		if(mode == CKNIFE_ENCRYPT || mode == CKNIFE_DECRYPT ||
				mode == CKNIFE_HMAC_SHA384) {
#if DEBUG > 0
			printf("\nHMAC Input: ");
			for(i=0; i<sizeof(in); i++) printf("%02x", data[i]);
#endif
			SHA384_Init(&c); /* inner */
			SHA384_Update(&c, ipad, sizeof(ipad));
			SHA384_Update(&c, data, sizeof(in));
			SHA384_Final(md, &c);
			/* md = SHA384(ipad, data) */
			SHA384_Init(&c); /* outer */
			SHA384_Update(&c, opad, sizeof(opad));
			SHA384_Update(&c, md, sizeof(opad));
			SHA384_Final(md, &c);
			/* md = SHA384(opad, SHA384(ipad, data)) */
#if DEBUG > 0
			printf("\nHMAC Output: ");
			for(i=0; i<sizeof(md); i++) printf("%02x", md[i]);
#endif
		}
		if(mode == CKNIFE_SHA384) {
			/* plain SHA384 */
			memcpy(md, in, sizeof(in));
		}

		/* add local MAC to the global MAC */
		SHA384_Update(&cglobal, md, sizeof(md));

	}

	desdone();
	memset(&ks1, 0, sizeof(ks1));
	memset(&ks2, 0, sizeof(ks2));
	memset(&ks3, 0, sizeof(ks3));
	memset(iwhite, 0, sizeof(iwhite));
	memset(owhite, 0, sizeof(owhite));
	close(fdin);

	/* get the global MAC */
	SHA384_Final(md, &cglobal);
#if DEBUG > 0
	printf("\nFinal MAC: ");
	for(i=0; i<MAC_SIZE; i++) printf("%02x", md[i]);
	printf("\n");
#endif

	/* print the resulting global MAC */
	if(mode == CKNIFE_HMAC_SHA384 || mode == CKNIFE_SHA384) {
		for(i=0; i<MAC_SIZE; i++) printf("%02x", md[i]);
		printf("\n");
	}
	if(mode == CKNIFE_DECRYPT) {
		/* verify resulting plaintext integrity */
		r = memcmp(md, resmd, MAC_SIZE);
		if(r != 0) {
			printf("error: plaintext integrity check failed!\n");
			printf("error: original mac:");
			for(i=0; i<MAC_SIZE; i++) printf("%02x", resmd[i]);
			printf("\nerror: plaintext mac:");
			for(i=0; i<MAC_SIZE; i++) printf("%02x", md[i]);
			printf("\n");
			close(fdout);
#if DEBUG > 0
			unlink(ofile);
#endif
			return 20;
		}

		return 0;
	}
	if( mode == CKNIFE_ENCRYPT) {
		/* store global MAC and length into the
		 * placeholders in output file
		 */
		lseek(fdout, 0, SEEK_SET);
		w = write(fdout, md, MAC_SIZE);
		if(w < 20) {
			perror("write MAC");
			return 21;
		}
		length = htonl(length);
		w = write(fdout, &length, sizeof(length));
		if( w < sizeof(length)) {
			perror("write length");
			return 22;
		}
	}

	close(fdout);

	return 0;

}
